#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <poll.h>
#include <errno.h>
#define __NO_STANDALONE_DBG_INC__
#include "platform.h"
#include "debuglib.h"
#include "di_defs.h"
#include "man_defs.h"
#include "debug_if.h"
#include "xdi_msg.h"
#include "dbgioctl.h"
#include <xdi_vers.h>
#include <string.h>

/*
  LOCALS
  */
#define DBG_MAGIC (0x47114711L)
#define DIVA_TRACE_HEADER_TITLE "\n# HEADER SUBJECT: "
#if !defined(DIVA_BUILD)
#define DIVA_BUILD "local"
#endif

typedef struct {
	const char*			key;
	const char*     text[64];
} diva_ditrace_options_t;

static void do_dprintf (char* fmt, ...);
static int ditrace_process_cmd_line (int argc, char** argv);
static int diva_decode_message (diva_dbg_entry_head_t* pmsg, int length);
static int diva_write_binary_ring_buffer (dword buffer_size, int do_poll);
static int diva_read_binary_ring_buffer (FILE* InStream);
static void usage(char *name);
static int add_file (byte** , const char* file, int limit);
static void fill_id_list (const char* id_array);
static int check_id (int id);

static char mnemonic[] = "LFETX   RMSITB  tncsxr  0123    ";
static FILE* OutStream;
static int DilogFormat = 0;
static int use_mlog_time_stamp = 0;
static int silent_operation = 0;

#define DIVA_MAINT_IDS 255
static int id_list[DIVA_MAINT_IDS];
static int max_id_index = 0;

/*
  IMPORTS
  */
extern FILE* VisualStream;
extern FILE* DebugStream;
extern void monitorLogs (void *inputFile, char *buf, unsigned long *bufSize) ;
extern void infoAnalyse (FILE *out, DbgMessage *msg) ;
extern void	pppAnalyse  (FILE *out, DbgMessage *msg) ;
extern void	tapiFilter  (FILE *out, DbgMessage *msg) ;
extern void dilogAnalyse (FILE *out, DbgMessage *msg, int DilogFormat) ;
extern void xlogAnalyse (FILE *out, DbgMessage *msg) ;
extern word xlog(byte *stream, void * buffer);


#define MAINT_DRIVER_NAME		"/dev/DivasMAINT"

typedef struct _diva_cmd_line_parameter {
	const char* ident;
	int found;
	int string;
	int value;
	int length;
	dword min;
	dword max;
	dword  vi;
	char vs[256];
} diva_cmd_line_parameter_t;

static diva_cmd_line_parameter_t ditrace_parameters[] = {
/*						              found
                            |  string
                            |  |  value
                            |  |  |     length
                            |  |  |     |  min
                            |  |  |     |  |           max
                            |  |  |     |  |           |  vi
                            |  |  |     |  |           |  |  vs */
{ "l",                      0, 0, 0,    0, 0,          0, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{ "p",                      0, 0, 0,    0, 0,          0, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{ "s",                      0, 0, 0,    0, 0,          0, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{ "d",                      0, 1, 1,  255, 0,          0, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{ "m",                      0, 0, 2,    0, 0, 0x7fffffff, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{ "Silent",                 0, 0, 0,    0, 0,          0, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{"File",                    0, 1, 1,  255, 0,          0, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{ "w",                      0, 0, 1,    0,64,   512*1024, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{"i",                       0, 1, 1,  255, 0,          0, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{ "-version",               0, 0, 0,    0, 0,          0, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{ "-help",                  0, 0, 0,    0, 0,          0, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{ "h",                      0, 0, 0,    0, 0,          0, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{ "?",                      0, 0, 0,    0, 0,          0, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{ "UseMlogTime",            0, 0, 0,    0, 0,          0, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{"s",                       0, 1, 1,  255, 0,          0, 0, {0, 0}},
/* ------------------------------------------------------------------------ */
{  0 , 0, 0, 0, 0, 0,    0, 0, {0, 0}}
};

static diva_cmd_line_parameter_t* ditrace_find_entry (const char* c) {
	int i;
	
	for (i = 0; ditrace_parameters[i].ident; i++) {
		if (!strcmp(ditrace_parameters[i].ident, c)) {
			return (&ditrace_parameters[i]);
		}
	}

	return (0);
}

static int ditrace_list_drivers (char* out, int length) {
  int i, fd =  open (MAINT_DRIVER_NAME, O_RDWR | O_NONBLOCK);
  byte data[512];
  dword* pcmd = (dword*)&data[0];
  dword hours, min, sec, msec;

  if (fd < 0) {
    if (!silent_operation)
      perror ("DITRACE");
    return (-1);
  }

  if (out && length > 1) {
    int i;
    char tmp[256];

    *out++ = '\n';
    length--;
    for (i = 0; ((i < 53) && (length > 1)); length--, i++) {
      *out++ = '-';
    }
    if (length > 1) {
      *out++ = '\n';
      length--;
    }
	  i = sprintf (tmp, "  ditrace, BUILD (%s[%s]-%s-%s)\n",
             DIVA_BUILD, diva_xdi_common_code_build, __DATE__, __TIME__);
    if (length > (i+1)) {
      strcpy (out, tmp);
      length-=i;
      out+= i;
		}
    for (i = 0; ((i < 53) && (length > 1)); length--, i++) {
      *out++ = '-';
    }
    if (length > 1) {
      *out++ = '\n';
      length--;
    }
    *out = 0;
  }
  if (!out) {
    fprintf (OutStream, "\n -----------------------------------------------------\n");
	  fprintf (OutStream,
             "  ditrace, BUILD (%s[%s]-%s-%s)\n",
             DIVA_BUILD, diva_xdi_common_code_build, __DATE__, __TIME__);
    fprintf (OutStream, " ------------------------------------------------------\n");
  }

  for (i = 1; ; i++) {
    pcmd[0]      = DITRACE_CMD_GET_DRIVER_INFO;
    pcmd[1]      = i; /* driver number */

    if (read (fd,
               pcmd,
               sizeof(data)) <= 0) {
      break;
    }
    if (!data[0]) {
      continue;
    }

    sec  = (data[1] | (data[2] << 8) | (data[3] << 16) | (data[4] << 24));
    msec = (data[5] | (data[6] << 8) | (data[7] << 16) | (data[8] << 24));

    hours = sec/3600;
    min   = (sec - hours*3600)/60;
    sec   = sec - hours*3600 - min*60;

    if (out) {
      char tmp [256];
      int len = sprintf (tmp, "  %2d:%02d:%02d.%03d %02d - %s\n",
                         hours, min, sec, msec, (byte)i, &data[9]);
      if (length >= (len+1)) {
        strcpy (out, tmp);
        length -= len;
        out    += len;
      }
    } else {
      fprintf (OutStream, "  %2d:%02d:%02d.%03d %02d - %s\n",
               hours, min, sec, msec, (byte)i, &data[9]);
    }
  }

  close (fd);

  if (out) {
    for (i = 0; ((i < 53) && (length > 1)); length--, i++) {
      *out++ = '-';
    }
    if (length > 1) {
      *out++ = '\n';
      length--;
    }
    *out = 0;
  } else {
    fprintf (OutStream, " ------------------------------------------------------\n");
  }

  return (0);
}

static int ditrace_get_driver_mask (int id) {
  int fd = open (MAINT_DRIVER_NAME, O_RDWR | O_NONBLOCK);
  byte data[4*sizeof(dword)];
  dword* pcmd = (dword*)&data[0];
  dword mask;

  if (fd < 0) {
    if (!silent_operation)
      perror ("DITRACE");
    return (-1);
  }

  pcmd[0] = DITRACE_READ_DRIVER_DBG_MASK;
  pcmd[1] = (dword)id;

  if (read (fd,
             pcmd,
             sizeof(data)) != 4) {
    if (!silent_operation)
      perror("DITRACE");
    close (fd);
    return (-1);
  }
  mask = data[0] | (data[1] << 8) | (data[2] << 16) | (data[3] << 24);
  fprintf (OutStream, "%08x\n", mask);

  close (fd);

  return (0);
}

static int ditrace_write_driver_mask (int id, dword mask) {
  int fd = open (MAINT_DRIVER_NAME, O_RDWR | O_NONBLOCK);
  byte data[4*sizeof(dword)];
  dword* pcmd = (dword*)&data[0];

  if (fd < 0) {
    if (!silent_operation)
      perror ("DITRACE");
    return (-1);
  }

  mask &= 0x7fffffff;

  pcmd[0] = DITRACE_WRITE_DRIVER_DBG_MASK;
  pcmd[1] = (dword)id;
  pcmd[2] = mask;

  if (write (fd,
             pcmd,
             sizeof(data)) != 4) {
    if (!silent_operation)
      perror("DITRACE");
    close (fd);
    return (-1);
  }

  close (fd);

  return (0);
}

int ditrace_read_traces (int do_poll) {
  int fd;
  diva_dbg_entry_head_t* pmsg;
  byte data[MSG_FRAME_MAX_SIZE+sizeof(*pmsg)];
  dword* pcmd = (dword*)&data[0];
	struct pollfd fds[2];
  int ret;

  if (ditrace_list_drivers (0,0)) {
    return (-1);
  }

  pmsg = (diva_dbg_entry_head_t*)data;

  if ((fd = open (MAINT_DRIVER_NAME, O_RDWR | O_NONBLOCK)) < 0) {
    if (!silent_operation)
      perror ("DITRACE");
    return (-1);
  }

  pcmd[0] = DITRACE_READ_TRACE_ENTRY;
  pcmd[1] = (dword)id_list[0];
  pcmd[2] = sizeof(data);

  for (;;) {
    while ((ret=read (fd,
                       pcmd,
                       sizeof(data))) > sizeof(*pmsg)) {
      if (diva_decode_message (pmsg, ret)) {
        close(fd);
        return (-1);
      }
      pcmd[0] = DITRACE_READ_TRACE_ENTRY;
      pcmd[1] = (dword)id_list[0];
      pcmd[2] = sizeof(data);
    }
    if (ret < 0) {
      if (!silent_operation)
        perror("DITRACE");
      close (fd);
      return (-1);
    }

    memset (&fds, 0x00, sizeof (fds));
    fds[0].fd     = fd;
    fds[0].events = POLLIN;
    if (!do_poll || (!(poll (fds, 1, -1) > 0)) ||
        (!(fds[0].revents & POLLIN))) {
        break;
    }
  }

  close (fd);

  return (0);
}

static void fill_id_list (const char* id_array) {
	int i = 0;
	const char* p = strstr (id_array, ",");

	memset (id_list, 0x00, sizeof(id_list));
	if ((id_list[i] = atoi(id_array))) {
		i++;
	}

	while (p && *p && p[1] && (i < DIVA_MAINT_IDS)) {
		p++;
		if ((id_list[i] = atoi(p))) {
			i++;
		}
		p = strstr (p, ",");
	}
	max_id_index = i;
}

static int check_id (int id) {
	int i;

	if (!(max_id_index && id)) {
		return (0);
	}

	for (i = 0; i < max_id_index; i++) {
		if (id == id_list[i]) {
			return (0);
		}
	}

	return (-1);
}

int ditrace_main (int argc, char** argv) {
  diva_cmd_line_parameter_t* e;

	my_dprintf = do_dprintf;
	VisualStream = stdout;
	DebugStream  = VisualStream;
  OutStream    = VisualStream;

  ditrace_process_cmd_line (argc, argv);

	if (ditrace_find_entry ("?")->found ||
	    ditrace_find_entry ("h")->found ||
	    ditrace_find_entry ("-help")->found) {
    usage (argv[0]);
    return (0);
  }

	if (ditrace_find_entry ("UseMlogTime")->found) {
    use_mlog_time_stamp = 1;
  }
	if (ditrace_find_entry ("Silent")->found) {
		if (!(VisualStream = fopen ("/dev/null", "w"))) {
			return (-1);
		}
    silent_operation = 1;
	  DebugStream  = VisualStream;
  }

	if (ditrace_find_entry ("-version")->found) {
		fprintf (VisualStream,
						 "ditrace, BUILD (%s[%s]-%s-%s)\n",
             DIVA_BUILD, diva_xdi_common_code_build, __DATE__, __TIME__);
		return (0);
	}
  if ((e = ditrace_find_entry ("File"))->found == 1) {
		if (!(OutStream = fopen (e->vs, "w+"))) {
      if (!silent_operation)
        perror ("Output File");
      return (-1);
    }
  }

  if (ditrace_find_entry ("l")->found) {
    return (ditrace_list_drivers(0,0));
  }

  if ((e = ditrace_find_entry ("d"))->found) {
		fill_id_list (e->vs);
  }
  if (ditrace_find_entry ("m")->found == 2) {
		int i, ret = -1;

		for (i = 0; i < DIVA_MAINT_IDS; i++) {
			if (id_list[i] && (ret = ditrace_get_driver_mask (id_list[i]))) {
				return (ret);
			}
		}

    return (ret);
  }
  if ((e = ditrace_find_entry ("m"))->found == 1) {
		int i, ret = -1;

		for (i = 0; i < DIVA_MAINT_IDS; i++) {
			if (id_list[i] && (ret = ditrace_write_driver_mask (id_list[i], e->vi))) {
				return (ret);
			}
		}

		return (ret);
  }

  if ((e = ditrace_find_entry ("w"))->found == 1) {
    /*
       Write plain binary data to the output file
       */
    if (!ditrace_find_entry ("File")->found) {
      fprintf (VisualStream, "%s\n", "'-File' missing");
      return (-1);
    }
    return (diva_write_binary_ring_buffer(e->vi, ditrace_find_entry("p")->found));
  }
  /*
    Read Binary message file
    */
  if ((e = ditrace_find_entry ("i"))->found == 1) {
    FILE* InStream = fopen (e->vs, "r+");

    if (!InStream) {
      if (!silent_operation)
        perror ("Input File");
      return (-1);
    }
    return (diva_read_binary_ring_buffer (InStream));
  }

  return (ditrace_read_traces(ditrace_find_entry("p")->found));
}

static int ditrace_process_cmd_line (int argc, char** argv) {
	int i;
	char* p;
	diva_cmd_line_parameter_t* e;
	diva_cmd_line_parameter_t* silent = ditrace_find_entry("Silent");
	

	for (i = 0; i < argc; i++) {
		p = argv[i];
		if (*p++ == '-') {
			if (!(e = ditrace_find_entry (p))) {
				if (!silent->found) {
					fprintf (VisualStream, "A: invalid parameter: (%s)\n", p);
				}
				exit (1);
			}
			if (e->found) {
				if (!silent->found) {
					fprintf (VisualStream, "A: dublicate parameter: (%s)\n", p);
				}
				exit (1);
			}

			e->found = 1;
			if (!e->value) {
				continue;
			}
			if (((i+1) >= argc) || (argv[i+1][0] == '-')) {
			  if (e->value == 2) {
          e->found = 2;
          continue;
        }
				if (!silent->found) {
					fprintf (VisualStream, "A: value missing: (%s)\n", e->ident);
				}
				exit (1);
			}
			p = &argv[i+1][0];
			if (e->string) {
				int len = strlen (p);
				if (len >= e->length) {
					if (!silent->found) {
						fprintf (VisualStream, "A: value too long: (%s)\n", e->ident);
					}
					exit (1);
				}
				strcpy (&e->vs[0], p);
			} else {
				e->vi = strtoul(p, 0, 0);
				if ((e->vi > e->max) || (e->vi < e->min)) {
					if (!silent->found) {
						fprintf (VisualStream, "A: invalid value (%s)\n", e->ident);
					}
					exit (1);
				}
			}
			i++;
		}
	}

	return (0);
}

static void do_dprintf (char* fmt, ...) {
	char buffer[256];
	va_list ap;

	buffer[0] = 0;

	va_start(ap, fmt);
	vsprintf (buffer, (char*)fmt, ap);
	va_end(ap);

	fprintf (DebugStream, "%s\n", buffer);
}

static int diva_decode_message (diva_dbg_entry_head_t* pmsg, int length) {
  byte* data = (byte*)&pmsg[1];
  dword DL_index, DT_index, DLI_name;
  static dword sequence = 0;
  dword hours, min, sec, msec;
  DbgMessage* dbg_msg;
  char msg_data [MSG_FRAME_MAX_SIZE+sizeof(*dbg_msg)+1];

  dbg_msg = (DbgMessage*)msg_data;
  DLI_name = DLI_NAME(pmsg->dli) ;	/* numeric debug level	*/
  DL_index = DL_INDEX(pmsg->dli) ;	/* position of mask bit */
  DT_index = DT_INDEX(pmsg->dli) ;	/* basic message type	*/

  if (pmsg->facility != MSG_TYPE_MLOG) {
    if (DL_index >= 32) {
      fprintf (DebugStream, "I: corrupted message, index=%d\n", DL_index);
      return (-1);
    }

    if (DLI_name == DLI_XLOG) {
      switch (data[1]) {
        case 0x01:	/* B-X */
          DL_index = DL_SEND;
          break;
        case 0x02:	/* B-R */
  				DL_index = DL_RECV;
          break;

        case 0x03:	/* D-X */
        case 0x04:	/* D-R */
        case 0x05:	/* SIG_EVENT */
        case 0x15:	/* xlog_sig */
        case 0x16:	/* event */
          DL_index = DL_CONN;
          break ;

        case 0x06:	/* low layer indications     */
        case 0x07:	/* low layer requests      s */
        case 0x0E:	/* network layer indications */
        case 0x0F:	/* network layer requests    */
          DL_index = DL_STAT;
          break;

        case 0x09:	/* MDL-ERROR */
          DL_index = DL_ERR;
          break ;

        case 0x14:	/* L1-Up, ... */
          DL_index = DL_LOG;
          break ;

        case 0x17:	/* EYE */
          DL_index = DL_PRV0;
          break;

        default:
          DL_index = DL_TRC ; break ;
      }
    }
  }

  switch (pmsg->facility) {
    case MSG_TYPE_DRV_ID:
    case MSG_TYPE_FLAGS:
      break;

    default:
      if (sequence > pmsg->sequence) {
        fprintf (OutStream, "sequence mismatch %d / %d\n",
                 sequence, pmsg->sequence);
      } else {
        if ((++sequence) < pmsg->sequence) {
          fprintf (OutStream, "==> missing sequence %d - %d\n",
                   sequence, pmsg->sequence - 1) ;
        }
      }
      sequence = pmsg->sequence;
  }

	if (check_id (pmsg->drv_id)) {
		return (0);
	}

  sec  = pmsg->time_sec;
  msec = pmsg->time_usec/1000;

  hours = sec/3600;
  min   = (sec - hours*3600)/60;
  sec   = sec - hours*3600 - min*60;



  if (pmsg->facility == MSG_TYPE_MLOG) {
    MI_XLOG_HDR *TrcData;
    byte tmp_xlog_buffer[4096*4];
    word tmp_xlog_pos;

    TrcData = (MI_XLOG_HDR *)&pmsg[1];

    if (use_mlog_time_stamp) {
      sec  = TrcData->time/1000;
      msec = TrcData->time - sec*1000;
      hours = sec/3600;
      min   = (sec - hours*3600)/60;
      sec   = sec - hours*3600 - min*60;
    }
    fprintf (OutStream, "%2d:%02d:%02d.%03d ", hours, min, sec, msec);
    fprintf (OutStream, "%c %X ", '-', (int)pmsg->drv_id) ;

    tmp_xlog_buffer[0] = 0;
    tmp_xlog_pos = xlog(&tmp_xlog_buffer[0], (void*)&TrcData->code);
    tmp_xlog_buffer[tmp_xlog_pos] = 0;
    fprintf (OutStream, "%s", &tmp_xlog_buffer[0]);
    fflush (OutStream);

    return (0);
  }

  /*
    Output time stamp
    */
  fprintf (OutStream, "%2d:%02d:%02d.%03d ", hours, min, sec, msec);

  /*
    append message type and driver id for regular messages
    */
  if ((pmsg->facility != MSG_TYPE_BINARY) && pmsg->drv_id &&
      (DLI_name != DLI_XLOG) && (DLI_name != DLI_MXLOG)) {
    fprintf (OutStream, "%c %X ", mnemonic[DL_index], (int)pmsg->drv_id) ;
  }

  /*
    Construct trace message and pass to decoder
    */
  dbg_msg->NTtime.LowPart	 = 0;
  dbg_msg->NTtime.HighPart = 0;
  dbg_msg->size	= (word)(length-sizeof(*pmsg));
  if (pmsg->facility == MSG_TYPE_STRING) {
    dbg_msg->size--;
  }
  dbg_msg->ffff	= 0xffff ;
  dbg_msg->id		= (word)pmsg->drv_id;
  dbg_msg->type	= (pmsg->dli | pmsg->facility);
  dbg_msg->seq	= pmsg->sequence;
  memcpy (&dbg_msg->data[0], data, dbg_msg->size);

  /*
    convert message body
    */
    switch (pmsg->facility) {
      case MSG_TYPE_STRING:
        tapiFilter (OutStream, dbg_msg) ;
        break;

       case MSG_TYPE_BINARY:
         switch (DLI_name) {
           case DLI_XLOG:
             dilogAnalyse (OutStream, dbg_msg, DilogFormat);
             break;
           case DLI_MXLOG:
             xlogAnalyse (OutStream, dbg_msg);
             break;
           case DLI_BLK:
             infoAnalyse (OutStream, dbg_msg);
             break ;
           default:
             pppAnalyse (OutStream, dbg_msg);
             break ;
         }
         break ;

      default:
        fprintf (OutStream,
                 "unknown message type 0x%x, fatal happens !\n",
                 DT_index) ;
  }

  fflush (OutStream) ;

  return (0);
}

typedef struct MSG_QUEUE {
	dword	Size;		/* total size of queue (constant)	*/
	byte	*Base;		/* lowest address (constant)		*/
	byte	*High;		/* Base + Size (constant)		*/
	byte	*Head;		/* first message in queue (if any)	*/
	byte	*Tail;		/* first free position			*/
	byte	*Wrap;		/* current wraparound position 		*/
	dword	Count;		/* current no of bytes in queue		*/
} MSG_QUEUE;

typedef struct MSG_HEAD {
	volatile dword	Size;		/* size of data following MSG_HEAD	*/
#define	MSG_INCOMPLETE	0x8000	/* ored to Size until queueCompleteMsg 	*/
} MSG_HEAD;

#define queueCompleteMsg(p) do{ ((MSG_HEAD *)p - 1)->Size &= ~MSG_INCOMPLETE; }while(0)
#define queueCount(q)	((q)->Count)
#define MSG_NEED(size) \
	( (sizeof(MSG_HEAD) + size + sizeof(dword) - 1) & ~(sizeof(dword) - 1) )

static void queueInit (MSG_QUEUE *Q, byte *Buffer, dword sizeBuffer) {
	Q->Size = sizeBuffer;
	Q->Base = Q->Head = Q->Tail = Buffer;
	Q->High = Buffer + sizeBuffer;
	Q->Wrap = 0;
	Q->Count= 0;
}

static byte *queueAllocMsg (MSG_QUEUE *Q, word size) {
	/* Allocate 'size' bytes at tail of queue which will be filled later
   * directly whith callers own message header info and/or message.
   * An 'alloced' message is marked incomplete by oring the 'Size' field
   * whith MSG_INCOMPLETE.
   * This must be reset via queueCompleteMsg() after the message is filled.
   * As long as a message is marked incomplete queuePeekMsg() will return
   * a 'queue empty' condition when it reaches such a message.  */

	MSG_HEAD *Msg;
	word need = MSG_NEED(size);

	if (Q->Tail == Q->Head) {
		if (Q->Wrap || need > Q->Size) {
			return(0); /* full */
		}
		goto alloc; /* empty */
	}
	
	if (Q->Tail > Q->Head) {
		if (Q->Tail + need <= Q->High) goto alloc; /* append */
		if (Q->Base + need > Q->Head) {
			return (0); /* too much */
		}
		/* wraparound the queue (but not the message) */
		Q->Wrap = Q->Tail;
		Q->Tail = Q->Base;
		goto alloc;
	}

	if (Q->Tail + need > Q->Head) {
		return (0); /* too much */
	}

alloc:
	Msg = (MSG_HEAD *)Q->Tail;

	Msg->Size = size | MSG_INCOMPLETE;

	Q->Tail	 += need;
	Q->Count += size;



	return ((byte*)(Msg + 1));
}

static void queueFreeMsg (MSG_QUEUE *Q) {
/* Free the message at head of queue */

	word size = ((MSG_HEAD *)Q->Head)->Size & ~MSG_INCOMPLETE;

	Q->Head  += MSG_NEED(size);
	Q->Count -= size;

	if (Q->Wrap) {
		if (Q->Head >= Q->Wrap) {
			Q->Head = Q->Base;
			Q->Wrap = 0;
		}
	} else if (Q->Head >= Q->Tail) {
		Q->Head = Q->Tail = Q->Base;
	}
}

static byte *queuePeekMsg (MSG_QUEUE *Q, word *size) {
	/* Show the first valid message in queue BUT DONT free the message.
   * After looking on the message contents it can be freed queueFreeMsg()
   * or simply remain in message queue.  */

	MSG_HEAD *Msg = (MSG_HEAD *)Q->Head;

	if (((byte *)Msg == Q->Tail && !Q->Wrap) ||
	    (Msg->Size & MSG_INCOMPLETE)) {
		return (0);
	} else {
		*size = Msg->Size;
		return ((byte *)(Msg + 1));
	}
}

static int diva_write_binary_ring_buffer (dword buffer_size, int do_poll) {
  int fd;
  int rb_fd = fileno (OutStream);
  byte* rb, *p, *map_base;
  MSG_QUEUE* dbg_queue;
  diva_dbg_entry_head_t* pmsg;
  byte data[10*(MSG_FRAME_MAX_SIZE+sizeof(*pmsg))+1];
  dword* pcmd = (dword*)&data[0];
	struct pollfd fds[2];
  int ret;
  int fret = 0;
  word size;
	char vtbl[] = " +*";
  dword events = 0;
  byte* pdata;
  dword cur_length;
	int   header_length = 0;

  if (buffer_size < 256) {
    buffer_size = 256;
  }
  buffer_size *= 1024;
  buffer_size += getpagesize();
  buffer_size = (buffer_size/getpagesize())*getpagesize();

  pmsg = (diva_dbg_entry_head_t*)data;

  pcmd[0] = DITRACE_READ_TRACE_ENTRYS;
  pcmd[1] = (dword)0;
  pcmd[2] = sizeof(data);

  if (rb_fd < 0) {
    if (!silent_operation)
      perror ("Output File");
    return (-1);
  }

  if (do_poll)
    fprintf (VisualStream, "==> Create log file ... \n");

  {
    dword i;
    char *p = malloc(getpagesize());
    memset (p, 0x00, getpagesize());

    for (i = 0; i < buffer_size/getpagesize(); i++) {
      if (write (rb_fd, p, getpagesize()) != getpagesize()) {
        if (!silent_operation)
          perror ("Output File");
        fclose (OutStream);
				free(p);
        return (-1);
      }
    }
    fflush (0);
		free(p);
  }

  if (!(rb = (byte*)mmap (0, buffer_size, PROT_READ | PROT_WRITE, MAP_SHARED, rb_fd, 0))) {
    if (!silent_operation)
      perror ("Output Buffer");
    fclose (OutStream);
    return (-1);
  }
  map_base = rb;

  /*
    Store Magic
    */
  *(dword*)rb  = (dword)DBG_MAGIC; /* Store Magic */
  rb += sizeof(dword);

  /*
    Create Message Buffer Info Header (up to 2048 Bytes)
    */
  *(dword*)rb = 2048; /* Extension Field Length */
  rb += sizeof(dword);
  if (ditrace_list_drivers (rb, 2048)) {
    munmap(map_base, buffer_size);
    fclose (OutStream);
    return (-1);
  }
  rb += 2048;

	header_length += add_file (&rb, DATADIR"/divas_cfg.rc",
														 buffer_size - sizeof(MSG_QUEUE) - 512 - 2048 - header_length);

  {
    int i;

    for (i = 0; i < 128; i++) {
      char name[sizeof("/var/log/divaxxx.log")+32];
      sprintf (name, "/var/log/diva%d.log", i);
      header_length += add_file (&rb, name,
																 buffer_size - sizeof(MSG_QUEUE) - 512 - 2048 - header_length);
    }
  }

	if ((buffer_size - sizeof(MSG_QUEUE) - 512 - 2048 - header_length) < (16*1024)) {
    munmap(map_base, buffer_size);
    fclose (OutStream);
    return (-1);
	}

  /*
    Terminate Extension List
    */
  *(dword*)rb = 0;
  rb += sizeof(dword);

  /*
    Store Ring Buffer Base
    */
  *(void**)rb  =  (void*)(rb+sizeof(void*)); /* Store Base  */
  rb += sizeof(void*);

  /*
    Create Ring Buffer
    */
  dbg_queue = (MSG_QUEUE*)rb;
  queueInit (dbg_queue, rb + sizeof(MSG_QUEUE),
						 buffer_size - sizeof(MSG_QUEUE) - 512 - 2048 - header_length);

  if ((fd = open (MAINT_DRIVER_NAME, O_RDWR | O_NONBLOCK)) < 0) {
    if (!silent_operation)
      perror ("MAINT");
    munmap(map_base, buffer_size);
    fclose (OutStream);
    return (-1);
  }

  if (do_poll)
    fprintf (VisualStream, "==> Read Message Stream, please press 'q' to exit\n");

  pcmd[0] = DITRACE_READ_TRACE_ENTRYS;
  pcmd[1] = (dword)0;
  pcmd[2] = sizeof(data);

  for (;;) {
    while ((ret=read (fd,
                       pcmd,
                       sizeof(data))) > sizeof(*pmsg)) {
      pdata = &data[0];
      for (; (cur_length = (pdata[0] | (pdata[1] << 8) | (pdata[2] << 16) | (pdata[3] << 24)));) {
        pdata += 4;
        while (!(p = (byte*)queueAllocMsg (dbg_queue, (word)cur_length))) {
          if ((p = (byte*)queuePeekMsg (dbg_queue, &size))) {
            queueFreeMsg (dbg_queue);
          } else {
            break;
          }
        }
        if (p) {
          dword to_flush = (cur_length/getpagesize())*getpagesize();
          memcpy (p, pdata, cur_length);
          queueCompleteMsg (p);

          if (cur_length > to_flush) {
            to_flush++;
          }
          msync ((void*)((((dword)p)/getpagesize())*getpagesize()), to_flush, MS_ASYNC);
          msync (map_base, getpagesize(), MS_ASYNC);
        }
        pdata += cur_length;
      }

      pcmd[0] = DITRACE_READ_TRACE_ENTRYS;
      pcmd[1] = (dword)0;
      pcmd[2] = sizeof(data);

      events++;
    }
    if (ret < 0) {
      if (!silent_operation)
        perror("DITRACE");
      close (fd);
      return (-1);
    }
    if (!silent_operation && do_poll) {
      fprintf (VisualStream, "%c %d\r", vtbl[events%3], events);
      fflush (VisualStream);
    }

    memset (&fds, 0x00, sizeof (fds));
    fds[0].fd     = fd;
    fds[0].events = POLLIN;
    if (!silent_operation) {
		  fds[1].fd     = STDIN_FILENO;
		  fds[1].events = POLLIN;
    }
    if (do_poll && (poll (fds, 1 + !silent_operation , -1) > 0)) {
      if (fds[1].revents & POLLIN) {
        char buffer[4];
        int length = read (STDIN_FILENO, buffer, sizeof(buffer)-1);
        if (length && (buffer[0] == 'q')) {
          break;
        }
      }
    } else {
      break;
    }
  }

  msync(map_base, buffer_size, MS_SYNC);
  munmap(map_base, buffer_size);
  close (fd);
  fclose (OutStream);

  return (fret);
}

static int diva_read_binary_ring_buffer (FILE* InStream) {
  int rb_fd = fileno (InStream);
  byte* rb, *map_base, *org_base;
  MSG_QUEUE* dbg_queue;
  diva_dbg_entry_head_t* pmsg;
  dword buffer_size;
  word size;
  int   header_number = 1, show_jump = 1;
  dword header_length;

  if (rb_fd < 0) {
    if (!silent_operation)
      perror ("Input  File");
    fclose (InStream);
    return (-1);
  }

  buffer_size = (dword)lseek(rb_fd, 0, SEEK_END);
  lseek (rb_fd, 0, SEEK_SET);

  if (buffer_size < 4096) {
    fprintf (VisualStream, "ERROR: Invalid File Format\n");
    fclose (InStream);
    return (-1);
  }

  if (!(rb = (byte*)mmap (0, buffer_size, PROT_READ | PROT_WRITE, MAP_PRIVATE, rb_fd, 0))) {
    if (!silent_operation)
      perror ("Input Buffer");
    fclose (InStream);
    return (-1);
  }
  map_base = rb;
  if (*(dword*)rb != (dword)DBG_MAGIC) {
    fprintf (VisualStream, "ERROR: Invalid File Format\n");
    munmap(map_base, buffer_size);
    fclose (InStream);
    return (-1);
  }
  rb += sizeof(dword);

  while ((header_length = *(dword*)rb)) {
    rb += sizeof(dword);

		header_number++;

		if ((*rb == 0x00) || (*rb == 0xff)) {
			/*
				Extension for the future binary headers
				*/
			fprintf (OutStream, "Ignore binary header Nr:%d\n", header_number-2);
		} else {
			if (header_number == 2) {
      fprintf (OutStream, "%s", (char*)rb);
    } else {
				if (show_jump) {
					show_jump = 0;
					fprintf (OutStream, "%s",
								 "\n\n# ================================================================== \n#\n");
					fprintf (OutStream, "%s",
						"# PLEASE GO DIRECTLY TO 'START TRACE DATA' IF THE SYSTEM DIAGNOSTIC\n");
					fprintf (OutStream, "%s", "# MESSAGES ARE NOT RELEVANT TO YOU.\n");
					fprintf (OutStream, "%s",
								 "#\n# ================================================================== \n\n");
				}
				fprintf (OutStream,
								 "\n# ==================== START TRACE HEADER NR:%d ==================== \n\n",
								 header_number-2);
				fprintf (OutStream, "%s", (char*)rb);
				fprintf (OutStream,
								 "\n# ====================  END TRACE HEADER NR:%d  ==================== \n\n",
								 header_number-2);
    }
    }

    rb += header_length;
  }
  rb += sizeof(dword);

	fprintf (OutStream,
					 "\n# ====================    START TRACE DATA     ==================== \n\n");

  if (!(org_base = *(byte**)rb)) {
    fprintf (VisualStream, "ERROR: Invalid File Format\n");
    munmap(map_base, buffer_size);
    fclose (InStream);
    return (-1);
  }
  rb += sizeof(void*);

  dbg_queue = (MSG_QUEUE*)rb;

  if (!(dbg_queue->Count && dbg_queue->Count)) {
    fprintf (VisualStream, "INFO: Empty File\n");
    munmap(map_base, buffer_size);
    fclose (InStream);
    return (0);
  }
  dbg_queue->Base = rb + ((byte*)(dbg_queue->Base) - (byte*)org_base);
  dbg_queue->High = rb + ((byte*)(dbg_queue->High) - (byte*)org_base);
  dbg_queue->Head = rb + ((byte*)(dbg_queue->Head) - (byte*)org_base);
  dbg_queue->Tail = rb + ((byte*)(dbg_queue->Tail) - (byte*)org_base);
  if (dbg_queue->Wrap) {
    dbg_queue->Wrap = rb + ((byte*)(dbg_queue->Wrap) - (byte*)org_base);
  }

  while ((pmsg = (diva_dbg_entry_head_t*)queuePeekMsg (dbg_queue, &size))) {
    diva_decode_message (pmsg, size);
    queueFreeMsg (dbg_queue);
  }

  munmap(map_base, buffer_size);
  fclose (InStream);

  return (0);
}

unsigned short xlog_stream (void* out, void* buffer) {
 byte tmp[16*1024];
  word length = xlog(&tmp[0], buffer);
  tmp[length] = 0;
  fprintf ((FILE*)out, "%s", tmp);

  return (length);
}

static diva_ditrace_options_t ditrace_options [] = {
{"-l",           {"list registered adapters/drivers",
                  0}
},
{"-d  [1...]",   {"specify driver for debug/condiguration",
                  0}
},
{"-p",           {"poll debug driver continuously",
                  0}
},
{"-m",           {"read driver debug mask",
                  0}
},
{"-m [0...0x7fffffff]",
                 {"set driver debug mask",
                  0}
},
{"-w [64...524288]",
                 {"store raw data stream in ring buffer",
                  "with specified size (in Kilobytes)",
                  "Suggested value: 256 KBytes",
                  "Can be used only with '-File' option",
                  0}
},
{"-i XXX",       {"interpret raw debug data from file 'XXX'",
                  0}
},
{"-h or -?",    {"help", 0}},
{"--help",      {"help", 0}},
{"--version",   {"build version", 0}},
{"-File",        {"write information to file instead of console",
                  "Filename should be provided",
                  0}
},
{"-o",           {"write information to file instead of console",
                  "Filename should be provided",
                  0}
},
{"-Silent",      {"silent operation mode, do not display messages",
                  0}
},
{ 0, {0}}
};

static void usage_option (const char* tab, const diva_ditrace_options_t* opt) {
	char buffer[2048];
	int s, j, i;

	strcpy (buffer, tab);
	strcat (buffer, opt->key);
	s = strlen (buffer);
	while (s++ < 22) {
		strcat (buffer, " ");
	}
	j = strlen (buffer) + 2;
	strcat (buffer, ": ");
	strcat (buffer, opt->text[0]);
	fprintf (VisualStream, "%s\n", buffer);

	for (i = 1; (opt->text[i] && (i < 64)); i++) {
		buffer[0] = 0;
		s = 0;
		while (s++ < j) {
			strcat (buffer, " ");
		}
		strcat (buffer, opt->text[i]);
		fprintf (VisualStream, "%s\n", buffer);
	}
}

static void usage_options (const char* tab) {
	const diva_ditrace_options_t* opt = &ditrace_options [0];

	fprintf (VisualStream, " options:\n");

	while (opt->key) {
		usage_option (tab, opt++);
	}
	fprintf (VisualStream, "\n");
}

static void usage(char *name) {
	char* p = strstr(name, "/");
	while (p) {
		name = p+1;
		p = strstr (name, "/");
	}

	fprintf(VisualStream, "Usage: %s options\n\n", name);
	usage_options ("  ");
}

static int add_file (byte** data, const char* file, int limit) {
	struct stat s;
	int fd;
	int name_len = strlen(file);
	int length = name_len + 3 + sizeof(dword) +
								sizeof(DIVA_TRACE_HEADER_TITLE)-1; /* 2*CR + \nSUBJECT: + terminating zero + size */
	byte* p = *data + sizeof(dword);

	if ((fd = open (file, O_RDONLY)) < 0) {
		return (0);
	}
	if (fstat (fd, &s) < 0) {
		close(fd);
		return(0);
	}

	length += s.st_size;
  length += getpagesize();
  length  = (length/getpagesize())*getpagesize();

	if (length < limit) {
		memcpy (p, DIVA_TRACE_HEADER_TITLE, sizeof(DIVA_TRACE_HEADER_TITLE)-1);
		p += sizeof(DIVA_TRACE_HEADER_TITLE) - 1;
		memcpy (p, file, name_len);
		p += name_len;
		*p++ = '\n';
		*p++ = '\n';
		if (read (fd, p, s.st_size) != s.st_size) {
			return (0);
		}
		p += s.st_size;
		*p = 0;

		*(dword*)*data = (dword)(length - sizeof(dword));

		*data += length;

		return (length);
	}

	return (0);
}

