#include "platform.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include "diva.h"
#include "pc_maint.h"
#include "xdi_msg.h"

#define DIVA_SDRAM_WRITE_BLOCK_SZ	(64*1024)

typedef int (*divas_get_card_properti_proc_t)(dword card, int fd, void* data);
typedef struct _diva_get_card_property_proc_t {
	const char* name;
	divas_get_card_properti_proc_t get_proc;
	int length;
} diva_get_card_property_proc_t;

static int divas_get_card_serial_number (dword card, int fd, void* data);
static int divas_get_card_xlog	(dword card, int fd, void* data);
static int divas_get_pci_hw_config(dword card, int fd, void* str);
static int divas_get_card_state	(dword card, int fd, void* data);

static diva_get_card_property_proc_t card_properties[] = {
{"serial number", divas_get_card_serial_number,		64},
{"xlog",					divas_get_card_xlog,					sizeof(struct mi_pc_maint)},
{"pci hw config",	divas_get_pci_hw_config,			8*11+5},
{"card state",		divas_get_card_state,					64},
{ 0,              0,                            0,}
};


dword
divas_open_driver (void)
{
	int fd =  open ("/dev/Divas", O_RDWR | O_NONBLOCK);

	if (fd < 0) {
		return (DIVA_INVALID_FILE_HANDLE);
	}

	return ((dword)fd);
}

int
divas_close_driver (dword handle)
{
	if (handle != DIVA_INVALID_FILE_HANDLE) {
		close ((int)handle);
		return (0);
	}
	return (-1);
}

int
divas_get_card (int card_number)
{
	dword handle = divas_open_driver ();
	diva_xdi_um_cfg_cmd_t msg;

	if (handle == DIVA_INVALID_FILE_HANDLE) {
		return (-1);
	}

	memset (&msg, 0x00, sizeof(msg));
	msg.adapter = (dword)card_number;
	msg.command = DIVA_XDI_UM_CMD_GET_CARD_ORDINAL;

	if (write ((int)handle, (void *) &msg, sizeof(msg)) != sizeof(msg)) {
		divas_close_driver (handle);
		return (-1);
	}
	memset (&msg, 0x00, sizeof(msg));

	if (read ((int)handle, (void *) &msg, sizeof(msg)) != sizeof(dword)) {
		divas_close_driver (handle);
		return (-1);
	}

	divas_close_driver (handle);

	return ((int)(*(dword*)&msg));
}

/* --------------------------------------------------------------------------
		Recovery different card propertoes
	 -------------------------------------------------------------------------- */

int
divas_get_card_properties (int card_number,
                           const char* name,
                           char* buffer,
                           int   size)
{
	int i, ret;
	dword handle;

	for (i = 0; card_properties[i].name; i++) {
		if (!strcmp (name, card_properties[i].name)) {
			if (card_properties[i].length > size) {
				return (-1);
			}
			if (!buffer) {
				return (card_properties[i].length);
			}
			if ((handle = divas_open_driver ()) == DIVA_INVALID_FILE_HANDLE) {
				return (-1);
			}

			ret = (*(card_properties[i].get_proc))(card_number, (int)handle, buffer);

			divas_close_driver (handle);
			return (ret);
		}
	}

	return (-1);
}

/* --------------------------------------------------------------------------
		Return serial number of the card as ASCII string
	 -------------------------------------------------------------------------- */
static int
divas_get_card_serial_number (dword card, int fd, void* data)
{
	diva_xdi_um_cfg_cmd_t msg;

	memset (&msg, 0x00, sizeof(msg));
	msg.adapter = (dword)card;
	msg.command = DIVA_XDI_UM_CMD_GET_SERIAL_NR;

	if (write (fd, (void *) &msg, sizeof(msg)) != sizeof(msg)) {
		return (-1);
	}
	memset (&msg, 0x00, sizeof(msg));

	if (read (fd, (void *) &msg, sizeof(msg)) != sizeof(dword)) {
		return (-1);
	}

	return (sprintf ((char*)data, "%d", *(dword*)&msg)+1);
}

static int
divas_get_pci_hw_config (dword card, int fd, void* str)
{
	diva_xdi_um_cfg_cmd_t msg;
	dword data[9];

	memset (&msg, 0x00, sizeof(msg));
	msg.adapter = (dword)card;
	msg.command = DIVA_XDI_UM_CMD_GET_PCI_HW_CONFIG;

	if (write (fd, (void *) &msg, sizeof(msg)) != sizeof(msg)) {
		return (-1);
	}
	memset (&data, 0x00, sizeof(data));

	if (read (fd, (void *) &data, sizeof(data)) != sizeof(data)) {
		return (-1);
	}

	return (sprintf ((char*)str,
		"0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%02x",
		data[0], data[1], data[2], data[3], data[4], data[5], data[6],
		data[7], data[8])+1);
}

/* --------------------------------------------------------------------------
		Issue one XLOG request and read one XLOG buffer from the card
	 -------------------------------------------------------------------------- */
static int
divas_get_card_xlog (dword card, int fd, void* data)
{
	diva_xdi_um_cfg_cmd_t msg;

	memset (&msg, 0x00, sizeof(msg));
	msg.adapter = (dword)card;
	msg.command = DIVA_XDI_UM_CMD_READ_XLOG_ENTRY;

	if (write (fd, (void *) &msg, sizeof(msg)) != sizeof(msg)) {
		return (-1);
	}
	memset (data, 0x00, sizeof(struct mi_pc_maint));

	if (read (fd, data, sizeof(struct mi_pc_maint)) != sizeof(struct mi_pc_maint)) {
		return (-1);
	}

	return (sizeof(struct mi_pc_maint));
}

void*
diva_os_malloc (unsigned long flags, unsigned long size)
{
	void* ret = malloc (size);

	if (ret) {
		memset (ret, 0x00, size);
	}

	return (ret);
}

void
diva_os_free (unsigned long flags, void* ptr)
{
	if (ptr) {
		free (ptr);
	}
}

/* --------------------------------------------------------------------------
		START BOOTLOADER
	 -------------------------------------------------------------------------- */
int
divas_start_bootloader (int card_number)
{
	dword handle = divas_open_driver ();
	diva_xdi_um_cfg_cmd_t msg;

	if (handle == DIVA_INVALID_FILE_HANDLE) {
		return (-1);
	}

	memset (&msg, 0x00, sizeof(msg));
	msg.adapter = (dword)card_number;
	msg.command = DIVA_XDI_UM_CMD_RESET_ADAPTER;

	if (write ((int)handle, (void *) &msg, sizeof(msg)) != sizeof(msg)) {
		divas_close_driver (handle);
		return (-1);
	}

	divas_close_driver (handle);

	return (0);
}

/* --------------------------------------------------------------------------
		WRITE CARD MEMORY
	 -------------------------------------------------------------------------- */
int
divas_ram_write (int card_number,
                 int ram_number,
                 const void* buffer,
                 dword offset,
                 dword length)
{
	dword handle = divas_open_driver ();
	char tmp[DIVA_SDRAM_WRITE_BLOCK_SZ+sizeof(diva_xdi_um_cfg_cmd_t)];
	diva_xdi_um_cfg_cmd_t* msg = (diva_xdi_um_cfg_cmd_t*)&tmp[0];
	byte* data = (byte*)&msg[1];
	const byte* src  = (const byte*)buffer;
	dword len;

	if (handle == DIVA_INVALID_FILE_HANDLE) {
		return (-1);
	}

	while (length) {
		len = MIN(DIVA_SDRAM_WRITE_BLOCK_SZ, length);

		memset (msg, 0x00, sizeof(*msg));
		msg->adapter = (dword)card_number;
		msg->command = DIVA_XDI_UM_CMD_WRITE_SDRAM_BLOCK;
		msg->command_data.write_sdram.ram_number  = ram_number;
		msg->command_data.write_sdram.offset	= offset;
		msg->command_data.write_sdram.length      = len;
		msg->data_length = len;

		memcpy (data, src, len);

	  if (write ((int)handle, (void *) msg,  sizeof(*msg)+len) != (sizeof(*msg)+len)) {
			divas_close_driver (handle);
			return (-1);
		}

		src		 += len;
		offset += len;
		length -= len;
	}


	divas_close_driver (handle);

	return (0);
}

/*
		Start card at specified address
	*/
int
divas_start_adapter (int card_number, dword start_address, dword features)
{
	dword handle = divas_open_driver ();
	diva_xdi_um_cfg_cmd_t msg;

	if (handle == DIVA_INVALID_FILE_HANDLE) {
		return (-1);
	}

	memset (&msg, 0x00, sizeof(msg));
	msg.adapter = (dword)card_number;
	msg.command = DIVA_XDI_UM_CMD_START_ADAPTER;
	msg.command_data.start.offset		= start_address;
	msg.command_data.start.features = features;

	if (write ((int)handle, (void *) &msg, sizeof(msg)) != sizeof(msg)) {
		divas_close_driver (handle);
		return (-1);
	}

	divas_close_driver (handle);

	return (0);
}

int
divas_stop_adapter (int card_number)
{
	dword handle = divas_open_driver ();
	diva_xdi_um_cfg_cmd_t msg;

	if (handle == DIVA_INVALID_FILE_HANDLE) {
		return (-1);
	}

	memset (&msg, 0x00, sizeof(msg));
	msg.adapter = (dword)card_number;
	msg.command = DIVA_XDI_UM_CMD_STOP_ADAPTER;

	if (write ((int)handle, (void *) &msg, sizeof(msg)) != sizeof(msg)) {
		divas_close_driver (handle);
		return (-1);
	}

	divas_close_driver (handle);

	return (0);
}

static int
divas_get_card_state (dword card, int fd, void* data)
{
	diva_xdi_um_cfg_cmd_t msg;
	const char* state;

	memset (&msg, 0x00, sizeof(msg));
	msg.adapter = (dword)card;
	msg.command = DIVA_XDI_UM_CMD_GET_CARD_STATE;

	if (write (fd, (void *) &msg, sizeof(msg)) != sizeof(msg)) {
		return (-1);
	}
	memset (&msg, 0x00, sizeof(msg));

	if (read (fd, (void *) &msg, sizeof(msg)) != sizeof(dword)) {
		return (-1);
	}
	switch (*(dword*)&msg) {
		case 0:
			state = "ready";
			break;

		case 1:
			state = "active";
			break;

		case 2:
			state = "trapped";
			break;

		case 3:
			state = "out of service";
			break;

		default:
			state = "unknown";
	}
	strcpy ((char*)data, state);

	return (strlen((char*)data)+1);
}

int
divas_write_fpga (int card_number,
                  int fpga_number,
                  const void* buffer,
                  dword length)
{
	dword handle = divas_open_driver ();
	char *tmp;
	diva_xdi_um_cfg_cmd_t* msg;

	if (handle == DIVA_INVALID_FILE_HANDLE) {
		return (-1);
	}
	if (!(tmp = malloc (sizeof(*msg)+length))) {
		divas_close_driver (handle);
		return (-1);
	}

	msg = (diva_xdi_um_cfg_cmd_t*)tmp;
	memset (msg, 0x00, sizeof(*msg));
	memcpy (&msg[1], buffer, length);

	msg->adapter = (dword)card_number;
	msg->command = DIVA_XDI_UM_CMD_WRITE_FPGA;
	msg->command_data.write_fpga.fpga_number  = fpga_number;
	msg->command_data.write_fpga.image_length = length;

	if (write ((int)handle, (void *) msg, sizeof(*msg)+length) != (sizeof(*msg)+length)) {
		free (tmp);
		divas_close_driver (handle);
		return (-1);
	}

	free (tmp);
	divas_close_driver (handle);

	return (0);
}

int
divas_ram_read (int card_number,
                int ram_number,
                void* buffer,
                dword offset,
                dword length)
{
	dword handle = divas_open_driver ();
	dword len;
	diva_xdi_um_cfg_cmd_t msg;
	byte* dst = (byte*)buffer;

	if (handle == DIVA_INVALID_FILE_HANDLE) {
		return (-1);
	}

	while (length) {
		len = MIN(DIVA_SDRAM_WRITE_BLOCK_SZ, length);

		memset (&msg, 0x00, sizeof(msg));
		msg.adapter = (dword)card_number;
		msg.command = DIVA_XDI_UM_CMD_READ_SDRAM;
		msg.command_data.read_sdram.ram_number  = ram_number;
		msg.command_data.read_sdram.offset			= offset;
		msg.command_data.read_sdram.length      = len;

    if (write ((int)handle, (void *) &msg, sizeof(msg)) != sizeof(msg)) {
			break;
		}

	  if (read ((int)handle, (void *) dst, len) != len) {
			break;
		}

		length -= len;
		offset += len;
		dst    += len;
	}

	divas_close_driver (handle);

	return ((length == 0) ? 0 : (-1));
}

/*
		Set untranslated protocol features
	*/
int
divas_set_protocol_features (int card_number, dword features)
{
	dword handle = divas_open_driver ();
	diva_xdi_um_cfg_cmd_t msg;

	if (handle == DIVA_INVALID_FILE_HANDLE) {
		return (-1);
	}

	memset (&msg, 0x00, sizeof(msg));
	msg.adapter = (dword)card_number;
	msg.command = DIVA_XDI_UM_CMD_SET_PROTOCOL_FEATURES;
	msg.command_data.features.features = features;

	if (write ((int)handle, (void *) &msg, sizeof(msg)) != sizeof(msg)) {
		divas_close_driver (handle);
		return (-1);
	}

	divas_close_driver (handle);

	return (0);
}

